@@ -1,5 +1,3 @@
-#!/usr/bin/env ruby
-
 require 'rake'
 require 'rake/clean'
 require 'rake/tasklib'
@@ -21,10 +19,10 @@ module Rake
     attr_accessor :tmp_dir
     attr_accessor :ext_dir
     attr_accessor :lib_dir
-    attr_accessor :platform
     attr_accessor :config_options
     attr_accessor :source_pattern
     attr_accessor :extra_options
+    attr_writer :platform
 
     def platform
       @platform ||= RUBY_PLATFORM
@@ -48,6 +46,7 @@ module Rake
 
     def define
       fail "Extension name must be provided." if @name.nil?
+      @name = @name.to_s
 
       define_compile_tasks
     end
