@@ -1,4 +1,4 @@
-#!/usr/bin/env ruby
+require "rbconfig"
 
 require 'rake/baseextensiontask'
 require "rubygems/package_task"
@@ -8,10 +8,9 @@ require "rubygems/package_task"
 
 module Rake
   class ExtensionTask < BaseExtensionTask
-
     attr_accessor :config_script
     attr_accessor :cross_compile
-    attr_accessor :cross_platform
+    attr_writer :cross_platform
     attr_writer :cross_config_options
     attr_accessor :no_native
     attr_accessor :config_includes
@@ -19,13 +18,14 @@ module Rake
     def init(name = nil, gem_spec = nil)
       super
       @config_script = 'extconf.rb'
-      @source_pattern = "*.c"
+      @source_pattern = "*.{c,cc,cpp}"
       @compiled_pattern = "*.{o,obj,so,bundle,dSYM}"
       @cross_compile = false
       @cross_config_options = []
       @cross_compiling = nil
       @no_native = false
       @config_includes = []
+      @ruby_versions_per_platform = {}
     end
 
     def cross_platform
@@ -111,13 +111,23 @@ Rerun `rake` under MRI Ruby 1.8.x/1.9.x to cross/native compile.
       # platform usage
       platf = for_platform || platform
 
+      binary_path = binary(platf)
+
       # lib_path
       lib_path = lib_dir
 
+      lib_binary_path = "#{lib_path}/#{binary_path}"
+      lib_binary_dir_path = File.dirname(lib_binary_path)
+
       # tmp_path
       tmp_path = "#{@tmp_dir}/#{platf}/#{@name}/#{ruby_ver}"
       stage_path = "#{@tmp_dir}/#{platf}/stage"
 
+      tmp_binary_path = "#{tmp_path}/#{binary_path}"
+      tmp_binary_dir_path = File.dirname(tmp_binary_path)
+      stage_binary_path = "#{stage_path}/#{lib_path}/#{binary_path}"
+      stage_binary_dir_path = File.dirname(stage_binary_path)
+
       # cleanup and clobbering
       CLEAN.include(tmp_path)
       CLEAN.include(stage_path)
@@ -126,17 +136,18 @@ Rerun `rake` under MRI Ruby 1.8.x/1.9.x to cross/native compile.
 
       # directories we need
       directory tmp_path
-      directory "#{stage_path}/#{lib_path}"
-      directory lib_dir
+      directory tmp_binary_dir_path
+      directory lib_binary_dir_path
+      directory stage_binary_dir_path
 
       # copy binary from temporary location to final lib
       # tmp/extension_name/extension_name.{so,bundle} => lib/
-      task "copy:#{@name}:#{platf}:#{ruby_ver}" => [lib_path, "#{tmp_path}/#{binary(platf)}"] do
-        install "#{tmp_path}/#{binary(platf)}", "#{lib_path}/#{binary(platf)}"
+      task "copy:#{@name}:#{platf}:#{ruby_ver}" => [lib_binary_dir_path, tmp_binary_path] do
+        install tmp_binary_path, "#{lib_path}/#{binary_path}"
       end
       # copy binary from temporary location to staging directory
-      task "copy:#{@name}:#{platf}:#{ruby_ver}" => ["#{stage_path}/#{lib_path}", "#{tmp_path}/#{binary(platf)}"] do
-        cp "#{tmp_path}/#{binary(platf)}", "#{stage_path}/#{lib_path}/#{binary(platf)}"
+      task "copy:#{@name}:#{platf}:#{ruby_ver}" => [stage_binary_dir_path, tmp_binary_path] do
+        cp tmp_binary_path, stage_binary_path
       end
 
       # copy other gem files to staging directory
@@ -144,7 +155,7 @@ Rerun `rake` under MRI Ruby 1.8.x/1.9.x to cross/native compile.
 
       # binary in temporary folder depends on makefile and source files
       # tmp/extension_name/extension_name.{so,bundle}
-      file "#{tmp_path}/#{binary(platf)}" => ["#{tmp_path}/Makefile"] + source_files do
+      file tmp_binary_path => [tmp_binary_dir_path, "#{tmp_path}/Makefile"] + source_files do
         jruby_compile_msg = <<-EOF
 Compiling a native C extension on JRuby. This is discouraged and a
 Java extension should be preferred.
@@ -153,6 +164,9 @@ Java extension should be preferred.
 
         chdir tmp_path do
           sh make
+          if binary_path != File.basename(binary_path)
+            cp File.basename(binary_path), binary_path
+          end
         end
       end
 
@@ -172,8 +186,8 @@ Java extension should be preferred.
         # now add the extconf script
         cmd << abs_extconf.relative_path_from(abs_tmp_path)
 
-        # rbconfig.rb will be present if we are cross compiling
-        if t.prerequisites.include?("#{tmp_path}/rbconfig.rb") then
+        # fake.rb will be present if we are cross compiling
+        if t.prerequisites.include?("#{tmp_path}/fake.rb") then
           options.push(*cross_config_options(platf))
         end
 
@@ -212,7 +226,7 @@ Java extension should be preferred.
       # platform matches the indicated one.
       if platf == RUBY_PLATFORM then
         # ensure file is always copied
-        file "#{lib_path}/#{binary(platf)}" => ["copy:#{name}:#{platf}:#{ruby_ver}"]
+        file "#{lib_path}/#{binary_path}" => ["copy:#{name}:#{platf}:#{ruby_ver}"]
 
         task "compile:#{@name}" => ["compile:#{@name}:#{platf}"]
         task "compile" => ["compile:#{platf}"]
@@ -223,7 +237,6 @@ Java extension should be preferred.
       platf = for_platform || platform
 
       # tmp_path
-      tmp_path = "#{@tmp_dir}/#{platf}/#{@name}/#{ruby_ver}"
       stage_path = "#{@tmp_dir}/#{platf}/stage"
 
       # lib_path
@@ -240,6 +253,16 @@ Java extension should be preferred.
           # adjust to specified platform
           spec.platform = Gem::Platform.new(platf)
 
+          # set ruby version constraints
+          ruby_versions = @ruby_versions_per_platform[platf] || []
+          sorted_ruby_versions = ruby_versions.sort_by do |ruby_version|
+            ruby_version.split(".").collect(&:to_i)
+          end
+          spec.required_ruby_version = [
+            ">= #{ruby_api_version(sorted_ruby_versions.first)}",
+            "< #{ruby_api_version(sorted_ruby_versions.last).succ}"
+          ]
+
           # clear the extensions defined in the specs
           spec.extensions.clear
 
@@ -259,12 +282,12 @@ Java extension should be preferred.
           callback.call(spec) if callback
 
           # Generate a package for this gem
-          pkg = Gem::PackageTask.new(spec) do |pkg|
-            pkg.need_zip = false
-            pkg.need_tar = false
+          pkg = Gem::PackageTask.new(spec) do |p|
+            p.need_zip = false
+            p.need_tar = false
             # Do not copy any files per PackageTask, because
             # we need the files from the staging directory
-            pkg.package_files.clear
+            p.package_files.clear
           end
 
           # copy other gem files to staging directory if added by the callback
@@ -295,7 +318,7 @@ Java extension should be preferred.
       end
 
       # add binaries to the dependency chain
-      task "native:#{@gem_spec.name}:#{platf}" => ["#{stage_path}/#{lib_dir}/#{binary(platf)}"]
+      task "native:#{@gem_spec.name}:#{platf}" => ["#{stage_path}/#{lib_path}/#{binary(platf)}"]
 
       # ensure the extension get copied
       unless Rake::Task.task_defined?("#{lib_path}/#{binary(platf)}") then
@@ -329,10 +352,14 @@ Java extension should be preferred.
 
         # tweak lib directory only when targeting multiple versions
         if multi then
-          version =~ /(\d+.\d+)/
+          version =~ /(\d+\.\d+)/
           @lib_dir = "#{@lib_dir}/#{$1}"
         end
 
+        # Update cross compiled platform/version combinations
+        ruby_versions = (@ruby_versions_per_platform[for_platform] ||= [])
+        ruby_versions << version
+
         define_cross_platform_tasks_with_version(for_platform, version)
 
         # restore lib_dir
@@ -368,39 +395,30 @@ Java extension should be preferred.
       # define compilation tasks for cross platform!
       define_compile_tasks(for_platform, ruby_ver)
 
-      # chain fake.rb, rbconfig.rb and mkmf.rb to Makefile generation
+      # chain fake.rb and mkmf.rb to Makefile generation
       file "#{tmp_path}/Makefile" => ["#{tmp_path}/fake.rb",
-                                      "#{tmp_path}/rbconfig.rb",
                                       "#{tmp_path}/mkmf.rb"]
 
-      # copy the file from the cross-ruby location
-      file "#{tmp_path}/rbconfig.rb" => [rbconfig_file] do |t|
+      # copy the rbconfig from the cross-ruby location and
+      # genearte fake.rb for different ruby versions
+      file "#{tmp_path}/fake.rb" => [rbconfig_file] do |t|
         File.open(t.name, 'w') do |f|
-          f.write "require 'fake.rb'\n\n"
+          f.write fake_rb(for_platform, ruby_ver)
           f.write File.read(t.prerequisites.first)
         end
       end
 
       # copy mkmf from cross-ruby location
       file "#{tmp_path}/mkmf.rb" => [mkmf_file] do |t|
-        cp t.prerequisites.first, t.name
-        if ruby_ver < "1.9" && "1.9" <= RUBY_VERSION
-          File.open(t.name, 'r+t') do |f|
-            content = f.read
+        File.open(t.name, 'w') do |f|
+          content = File.read(t.prerequisites.first)
+          content.sub!(/^(require ')rbconfig(')$/, '\\1fake\\2')
+          if ruby_ver < "1.9" && "1.9" <= RUBY_VERSION
             content.sub!(/^(      break )\*(defaults)$/, '\\1\\2.first')
             content.sub!(/^(    return )\*(defaults)$/, '\\1\\2.first')
             content.sub!(/^(  mfile\.)print( configuration\(srcprefix\))$/, '\\1puts\\2')
-            f.rewind
-            f.write content
-            f.truncate(f.tell)
           end
-        end
-      end
-
-      # genearte fake.rb for different ruby versions
-      file "#{tmp_path}/fake.rb" do |t|
-        File.open(t.name, 'w') do |f|
-          f.write fake_rb(for_platform, ruby_ver)
+          f.write content
         end
       end
 
@@ -455,9 +473,7 @@ Java extension should be preferred.
           if RUBY_PLATFORM =~ /mswin/ then
             'nmake'
           else
-            ENV['MAKE'] || %w[gmake make].find { |c|
-              system("#{c} -v >> #{dev_null} 2>&1")
-            }
+            ENV['MAKE'] || find_make
           end
       end
 
@@ -468,8 +484,19 @@ Java extension should be preferred.
       @make
     end
 
-    def dev_null
-      windows? ? 'NUL' : '/dev/null'
+    def find_make
+      candidates = ["gmake", "make"]
+      paths = (ENV["PATH"] || "").split(File::PATH_SEPARATOR)
+
+      exeext = RbConfig::CONFIG["EXEEXT"]
+      candidates.each do |candidate|
+        paths.each do |path|
+          make = File.join(path, "#{candidate}#{exeext}")
+          return make if File.executable?(make)
+        end
+      end
+
+      nil
     end
 
     def compiled_files
@@ -484,14 +511,20 @@ Java extension should be preferred.
       [*@cross_platform].map { |p| "native:#{p}" }
     end
 
+    def ruby_api_version(ruby_version)
+      ruby_version.split(".")[0, 2].join(".")
+    end
+
     def fake_rb(platform, version)
       <<-FAKE_RB
         # Pre-load resolver library before faking, in order to avoid error
         # "cannot load such file -- win32/resolv" when it is required later on.
         # See also: https://github.com/tjschuck/rake-compiler-dev-box/issues/5
         require 'resolv'
+        require 'rbconfig'
 
         class Object
+          remove_const :RbConfig
           remove_const :RUBY_PLATFORM
           remove_const :RUBY_VERSION
           remove_const :RUBY_DESCRIPTION if defined?(RUBY_DESCRIPTION)
